#!/usr/bin/env python
#
# $Id: DarcHost.py 47 1996-04-27 20:56:08Z darc $
# 
# Jacob Martinson <martinson.jacob at gmail.com>
#
# This software is licensed under the Gnu Public License
# For more info see: 
#         http://www.gnu.org/licenses/gpl.html

import time, md5, re, random, popen2, os, threading, sys


# used to raise exceptions regarding operational issues like 
# sftp reporting permission denied, etc.
class OperationalError(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)


# used to copy aide stdout to local file without blocking
# and to allow us to stop if aide takes too long or hangs
class db_writer(threading.Thread):

	def __init__(self,popen_fh,db_path):
		threading.Thread.__init__(self)
		self.popen_fh = popen_fh
		self.db_path = db_path

	def run(self):

		if self.db_path == '__STDOUT__':
			db_fh = sys.stderr
		else:
			db_fh = open(self.db_path,'w')

		ln = self.popen_fh.readline()
		while ln:

			# if we are running in test mode, do not filter any output,
			# so error messages reported on remote end will make it back 
			# to the user
			if self.db_path == '__STDOUT__':
				db_fh.write(ln)
				db_fh.flush()

			# filter out the aide stderr output that gets mixed over ssh
			elif ln[0] == '/' or ln[0] == '@':
				db_fh.write(ln)
				db_fh.flush()
			ln = self.popen_fh.readline()


class host:
	def __init__(self,hostname,aconf,arch='',ipaddr='',user='',key='',port=22,connect_timeout='',remotedir=''):

		# symbolic hostname
		self.hostname = hostname         

		# if specified we use it instead of resolving hostname
		self.ipaddr = ipaddr             

		# ssh port number
		self.port   = port

		# system architecture for determining which binary to send
		self.arch   = arch

		# username to connect with
		self.user   = user

		# private key for authentication
		self.key    = key

		# gets passed with -oConnectTimeout to ssh and sftp
		self.connect_timeout = connect_timeout

		# config module
		self.config = None

		# aide config
		self.aconf = aconf

		# where to put aide binary on remote host
		self.remotedir = remotedir

		# status 
		#
		# PROB    - there was a problem checking the host, unable to report
		# PASS    - host was checked and aide reported no changes
		# FAIL    - host was checked and aide reported changes
		self.status = None

		# used to store information about why check was unable to be performed
		self.status_detail = []


		# time values, used for setting some directory paths
		self.year = time.strftime('%Y')
		self.month = time.strftime('%m')
		self.day = time.strftime('%d')


	def main(self,init=False):


		try:
			self.build_vars_from_config()

			if self.config.verbose:
				sys.stderr.write(self.vpref + 'START\n')

			self.store_binary()
	
			if not os.path.exists(self.baseline):
				init = True
				self.init = True

			self.run_aide(init=init)

			self.remove_binary()

			if not init:
				self.gen_report()
				self.parse_report()
				self.compress_archive()
		except OperationalError, inst:
			if self.config.verbose:
				sys.stderr.write('***DarcHost.main() caught OperationalError %s\n' % str(inst))
			self.status = 'PROB'
			self.status_detail.append(str(inst))
			if init:
				msg = str(inst)
				msg += '\n'
				sys.stderr.write(msg)
				sys.exit(2)
			return


	def compress_archive(self):
		cmd = 'gzip %s' % self.newdb
		os.system(cmd)


	def test(self):
		try:
			self.build_vars_from_config()
			self.store_binary()
			self.run_aide(test=True)
			self.remove_binary()

		except OperationalError, inst:
			if self.config.verbose:
				sys.stderr.write(self.vpref + '***DarcHost.test() caught OperationalError \n')
			sys.stderr.write(str(inst) + '\n')
			return
	

	def gen_report(self):

		if self.config.verbose: 
			sys.stderr.write(self.vpref + 'GEN_REPORT() - comparing to baseline\n')

		cmd = 'nice -n 19 %s --compare --config="-" ' % self.config.local_aide_binary
		proc = popen2.Popen4(cmd)

		config_fh = open(self.aconf,'r')

		report_fh = open(self.report,'w')

		proc.tochild.write('database=file://%s\n' % self.baseline)
		proc.tochild.write('database_new=file://%s\n' % self.newdb)

		config = config_fh.readlines()

		if self.config.verbose: 
			sys.stderr.write(self.vpref + ' - writing config to child\n')

		for ln in config:
			proc.tochild.write(ln)
		proc.tochild.close()

		if self.config.verbose: 
			sys.stderr.write(self.vpref + ' - writing report from child to disk:\n')
			sys.stderr.write(self.vpref + ' -   %s\n' % self.report)

		# read report from aide's stdout
		rln = proc.fromchild.readline()
		while rln:
			report_fh.write(rln)
			rln = proc.fromchild.readline()

		report_fh.close()

		code = proc.wait()
		if code != 0:
			if self.config.verbose:
				sys.stderr.write(self.vpref + ' - aide compare exited non-zero\n')
			report_fh = open(self.report,'r')
			self.status_detail.append(report_fh.readlines())
			raise OperationalError, 'Error running aide in compare mode'


	def parse_report(self):
		'Parses report file, sets self.status to indicate results'
		
		report_fh = open(self.report,'r')

		ln = report_fh.readline()
		if ln[0:22] == 'AIDE found differences':
			self.status = 'FAIL'
		else:
			self.status = 'PASS'

		report_fh.close()




	def build_vars_from_config(self):

		# This method sets a number of variables based on the contents of the 
		# config file, either setting default values specified there or setting
		# paths, etc.

		config = self.config

		# timestamp used in new database filename
		self.ts = time.strftime('%Y%m%d_%H.%M.%S')

		# location we expect to find the baseline database
		self.baseline = '%s/%s.db' % (config.dbdir,self.hostname)

		# location we write the newly generated database
		self.newdb    = '%s/%s.%s.db' % (config.dbarchdir,self.hostname,self.ts)

		# location we capture the aide report to
		self.report   = '%s/%s.%s.rep' % (config.reportdir,self.hostname,self.ts)

		# grab default values from config if none were passed to __init__()
		if self.user == '':
			self.user = config.default_user

		if self.arch == '':
			self.arch = config.default_arch

		if self.key == '':
			self.key  = config.default_key

		if self.remotedir == '':
			self.remotedir  = config.remotedir

		# which binary to send
		self.bin      = '%s/%s' % (config.bindir,config.binmap[self.arch])

		# make sure local copy of binary exists 
		if not os.path.isfile(self.bin):
			raise OperationalError, "This host's AIDE binary does not exist: %s" % self.bin
	
		# hname holds the hostname/ip address string we pass to ssh and sftp. 
		# If ipaddr is set we use that, otherwise let ssh/sftp resolve the 
		# hostname with the local resolver
		if self.ipaddr != '':
			self.hname = self.ipaddr
		else:
			self.hname = self.hostname

		if config.aide_filename == 'random':
			m = md5.new()
			m.update(str(random.random()))
			self.remote_bin = '%s/%s' % (self.remotedir,m.hexdigest()[0:8])
		else:
			self.remote_bin = '%s/%s' % (self.remotedir,config.aide_filename)

		self.aconf = '%s/%s' % (config.confdir,self.aconf)

		self.vpref = 'THREAD:%-12s: ' % self.hostname

		self.init = False


	def print_vars(self):
		c1 = 15
		print 'hostname:'.rjust(c1), self.hostname
		print 'baseline:'.rjust(c1), self.baseline
		print 'newdb:'.rjust(c1), self.newdb
		print 'user:'.rjust(c1), self.user
		print 'bin:'.rjust(c1), self.bin
		print 'key:'.rjust(c1), self.key


	def has_baseline(self):

		if os.path.exists(self.baseline):
			return True
		else:
			return False


	def store_binary(self):
		# Copies binary to host by sftp, returns True if successful, 
		# raises OperationalError otherwise.

		t1 = time.time()

		connect_timeout = self.config.connect_timeout
		sftp_xfer_timeout = self.config.sftp_xfer_timeout

		cmd = 'sftp -oIdentityFile="%s" -oConnectTimeout=%s -oPort=%s %s@%s' 
		cmd = cmd % (self.key,connect_timeout,self.port,self.user,self.hostname)
		if self.config.verbose:
			sys.stderr.write(self.vpref + cmd + '\n')
		p = popen2.Popen4(cmd)
		sc = 'put %s %s\n' % (self.bin,self.remote_bin)
		if self.config.verbose:
			sys.stderr.write(self.vpref + sc)
		p.tochild.write(sc)
		sc = 'chmod 700 %s\n' % self.remote_bin
		if self.config.verbose:
			sys.stderr.write(self.vpref + sc)
		p.tochild.write(sc)
		p.tochild.close()

		sleep_time = 0.1
		procrunning = True
		t2 = time.time()
		pollcode = p.poll()
		while (t2 - t1) < sftp_xfer_timeout and procrunning:
			if pollcode == -1:
				time.sleep(sleep_time)
				t2 = time.time()
				pollcode = p.poll()
			else:
				procrunning = False

		# sftp_xfer time has passed, kill proc and return failure code
		if procrunning:
			if self.config.verbose:
				sys.stderr.write(self.vpref + 'sftp_xfer time passed, killing pid %s\n' % p.pid)
			os.kill(p.pid,15)
			raise OperationalError, 'sftp_xfer_timeout passed - %s sec' % sftp_xfer_timeout

		# process exited before timeout, still need to check return code
		# and make sure there were no transfer errors reported 
		else:
			if self.config.verbose:
				sys.stderr.write(self.vpref + 'sftp exited code = %s\n' % pollcode)
			if pollcode != 0:
				raise OperationalError, 'sftp exited w non-zero code: %s' % pollcode
			else:

				# check sftp output for error messages
				lns = p.fromchild.readlines()
				err = False
				for ln in lns:
					if ln[0:20] == "Couldn't get handle:" or ln[-10:] == 'not found.':
						err = True
						errmsg = ln
				if err:
					raise OperationalError, 'sftp error- %s' % errmsg
				else:
					return True



	def remove_binary(self):
		# Removes binary from host by sftp, returns True if successful, 
		# raises OperationalError otherwise.

		t1 = time.time()

		connect_timeout = self.config.connect_timeout
		sftp_xfer_timeout = self.config.sftp_xfer_timeout
	
		cmd = 'sftp -oIdentityFile="%s" -oConnectTimeout=%s -oPort=%s %s@%s' 
		cmd = cmd % (self.key,connect_timeout,self.port,self.user,self.hostname)
		if self.config.verbose:
			sys.stderr.write(self.vpref + cmd + '\n')
		p = popen2.Popen4(cmd)
		sc = 'rm %s\n' % (self.remote_bin)
		if self.config.verbose:
			sys.stderr.write(self.vpref + sc)
		p.tochild.write(sc)
		p.tochild.close()

		sleep_time = 0.1
		procrunning = True
		t2 = time.time()
		pollcode = p.poll()
		while (t2 - t1) < sftp_xfer_timeout and procrunning:
			if pollcode == -1:
				time.sleep(sleep_time)
				t2 = time.time()
				pollcode = p.poll()
			else:
				procrunning = False

		# sftp_xfer time has passed, kill proc and return failure code
		if procrunning:
			if self.config.verbose:
				sys.stderr.write(self.vpref + 'sftp_xfer time passed, killing pid %s\n' % p.pid)
			os.kill(p.pid,15)
			raise OperationalError, 'sftp_xfer_timeout passed - %s sec' % sftp_xfer_timeout

		# process exited before timeout, still need to check return code
		# and make sure there were no transfer errors reported 
		else:
			if self.config.verbose:
				sys.stderr.write(self.vpref + 'sftp exited code = %s\n' % pollcode)
			if pollcode != 0:
				raise OperationalError, 'sftp exited w non-zero code: %s' % pollcode
			else:

				# check sftp output for error messages
				lns = p.fromchild.readlines()
				err = False
				for ln in lns:
					if ln[0:20] == "Couldn't get handle:" or ln[-10:] == 'not found.':
						err = True
						errmsg = ln
				if err:
					raise OperationalError, 'sftp error- %s' % errmsg
				else:
					return True


	def last_line_is_present(self):
		'Looks for @@end_db at end of db file to verify if it is complete.'

		if self.config.verbose:
			sys.stderr.write(self.vpref + 'looking for terminator in %s\n' % self.dbfile)

		fh = open(self.dbfile,'r')
		lastln = ''
		ln = fh.readline()
		while ln:
			lastln = ln
			ln = fh.readline()
		if lastln[0:8] == '@@end_db':
			return True
		else:
			return False




	def run_aide(self,init=False,test=False):

		config = self.config

		aide_timeout = config.aide_timeout

		# build ssh command
		cmd  = 'ssh -t -i %s -p %s -oConnectTimeout=%s '
		cmd  = cmd % (self.key,self.port,self.config.connect_timeout)
		cmd += '%s@%s ' % (self.user,self.hname)
		cmd += 'sudo nice -n 19 %s -i --config="-"' % self.remote_bin

		try:
			conf_fh = open(self.aconf,'r')
		except IOError, (errno,strerr):
			msg = 'Unable to open AIDE config, %s: ' % self.aconf
			msg += strerr
			raise OperationalError, msg

		conf_lns = conf_fh.readlines()
		conf_fh.close()

		if self.config.verbose:
			sys.stderr.write(self.vpref + cmd + '\n')

		t1 = time.time()

		p = popen2.Popen4(cmd)
		p.tochild.write('database_out=stdout\n')
		p.tochild.write('verbose=4\n')

		# if in test mode, just report on =/ to send a small 
		# database back to the user
		if test:
			p.tochild.write('=/ R\n')
		else:
			for ln in conf_lns:
				p.tochild.write(ln)
		p.tochild.close()

		# if in init mode, the db_writer instance needs to send the file
		# to the baseline location instead of newdb
		if init:
			dbfile = self.baseline
		elif test:
			if config.verbose:
				sys.stderr.write(self.vpref + 'setting database file to stdout\n')
			dbfile = '__STDOUT__'
		else:
			dbfile = self.newdb

		# set dbfile in self so we can check it later for completeness
		self.dbfile = dbfile

		# spawn a thread to read stdout without blocking
		if config.verbose:
			sys.stderr.write(self.vpref + 'starting db_writer(child,%s)\n' % dbfile)
		dbw = db_writer(p.fromchild,dbfile)
		dbw.start()


		sleep_time = 1
		procrunning = True
		t2 = time.time()
		pollcode = p.poll()
		while (t2 - t1) < aide_timeout and procrunning:
			if pollcode == -1:
				time.sleep(sleep_time)
				t2 = time.time()
				pollcode = p.poll()
			else:
				procrunning = False


		# aide_timeout has passed, kill proc and return failure code
		if procrunning:
			if self.config.verbose:
				sys.stderr.write(self.vpref + 'aide time passed, killing pid %s for host\n' % (p.pid,self.hostname))
			try:
				os.kill(p.pid,15)
				os.unlink(self.newdb)
			except Exception:
				pass
			raise OperationalError, 'aide_timeout passed - %s sec' % aide_timeout

		# process exited before timeout, still need to check return code
		# and make sure there were no transfer errors reported 
		else:

			if self.config.verbose:
				sys.stderr.write(self.vpref + 'ssh exited code = %s\n' % pollcode)

			if not test:
				if self.last_line_is_present():
	
					# check output for error messages
					lns = p.fromchild.readlines()
					err = False
					for ln in lns:
						# TODO - update this to check for ssh-specific error msgs
						if ln[0:20] == "Couldn't get handle:" or ln[-10:] == 'not found':
							err = True
							errmsg = ln
					if err:
						raise OperationalError, 'ssh error- %s' % errmsg
					else:
	
						# wait for the db_writer to read all the db output and write
						# it out to newdb
						dbw.join()
	
						if self.init:
							self.status = 'INIT'
	
						return True
	
				else:
	
					# put any error messages in self.status_detail for later reporting
					lns = p.fromchild.readlines()
					for ln in lns:
						self.status_detail.append(ln)
	
					raise OperationalError, 'database is missing @@end_db terminator for host %s' % self.hostname
