#!/usr/bin/env python
#
# $Id: darc.py 47 1996-04-27 20:56:08Z darc $
# 
# Jacob Martinson <martinson.jacob at gmail.com>
#
# The project page for this software is located here:
# http://icculus.org/projects/darc/
#
# This software is licensed under the Gnu Public License
# For more info see: 
#         http://www.gnu.org/licenses/gpl.html


import os.path, os, optparse, sys, re, popen2, imp, time, random, md5, threading, syslog

import DarcReport


class HostThread(threading.Thread):

	# This is a threaded wrapper around the host methods so they can
	# all run in parallel.  Timeout management is done within the 
	# DarcHost.host class methods, so we expect h.main() to return
	# in a timely manner.

	def __init__(self,h):
		threading.Thread.__init__(self)
		self.h = h
	
	def run(self):
		self.h.main()



def parse_opts():
	global args, options

	u = "usage: %prog -c <config file> "
	p = optparse.OptionParser(usage=u)

	p.add_option("-c","--config", dest="config", help="Path to config file")

	help = "Hostname to initialize.  Aide is run in initialize mode on the host and a baseline database is created.  All other hosts in the config are ignored."
	p.add_option("-i","--init", dest="inithost", help=help)

	help = "Hostname to test.  Aide is run in initialize mode on the host with the new database written to your STDOUT.  This option is useful to test new host setups.  It tests sftp and ssh connectivity, remote authentication, and the sudo setup on the remote host.  If you see a new aide database covering '/tmp R', your setup is good.  If you see error messages instead, your setup has problems."
	p.add_option("-t","--test", dest="testhost", help=help)

	help = "Update baseline database with the contents of the most recent archive database (based on timestamped directory and filename paths).  Used when you want to acknowledge and accept the filesystem changes in the last report for a particular host."
	p.add_option("-u","--update", dest="update_host", help=help)

	help = "Single host check.  Run normal parallel checks and generate report, but only for this one host."
	p.add_option("-s","--single", dest="single_host", help=help)

	help = "If logfile is specified, we will redirect STDOUT and STDERR to this logfile."
	p.add_option("-l","--logfile", dest="logfile", help=help)

	help = "Send a copy of the HTML report as an email attachment to each of the recipients listed in the config file."
	p.add_option("-e",action="store_true",dest="report_by_email",help=help)

	help = "Create the directory structure required by the settings defined in the config file."
	p.add_option("-k","--mkdirs", action="store_true",default=False,dest="mkdirs", help=help)

	help = "Log a message for each failed host to syslog."
	p.add_option("-y","--syslog", action="store_true",default=False,dest="syslog", help=help)

	help = "Verbose output, mostly for debugging and development purposes."
	p.add_option("-v","--verbose", action="store_true",default=False,dest="verbose", help=help)

	(options,args) = p.parse_args()

	if options.config == None:
		p.print_help()
		sys.exit(2)

	if options.testhost and options.inithost:
		msg  = 'Test (-t, --test) and initialize (-i, --init) options are mutually exclusive\n'
		print msg
		p.print_help()
		sys.exit(2)

	if options.mkdirs and (options.testhost or options.inithost):
		msg  = 'mkdirs is mutually exclusive with test (-t, --test) and initialize (-i, --init)\n'
		print msg
		p.print_help()
		sys.exit(2)



def check_directories():
	# verify expected directory structure exists, error out if it doesn't

	basedir   = config.basedir
	dbdir     = config.dbdir
	dbarchdir = config.dbarchdir
	tmpdir    = config.tmpdir
	bindir    = config.bindir
	reportdir = config.reportdir
	confdir   = config.confdir

	if options.verbose:
		sys.stderr.write("check_directories()\n")

	for d in [basedir,dbdir,dbarchdir,tmpdir,bindir,confdir,reportdir]:
		if not os.path.isdir(d):
			msg = '''Specified directory does not exist: %s
All directories defined in config can be created automatically with the 
-k or --mkdirs options.
''' % d
			sys.stderr.write(msg)
			remove_lockfile()
			sys.exit(2)

	local_bin = config.local_aide_binary
	if not os.path.isfile(local_bin):
		msg = 'Local aide binary for compare mode does not exist: %s\n' % local_bin
		sys.stderr.write(msg)
		sys.exit(2)


def add_dir_timestamps():
	'updates certain directory paths with timestamp'

	year  = time.strftime('%Y')
	month = time.strftime('%m')
	day   = time.strftime('%d')

	config.reportdirbase = config.reportdir

	config.dbarchdir = '%s/%s-%s/%s' % (config.dbarchdir,year,month,day)
	config.reportdir = '%s/%s-%s/%s' % (config.reportdir,year,month,day)


def mkdirs():
	# make directories specified in config, gets run with -k or --mkdirs

	basedir   = config.basedir
	dbdir     = config.dbdir
	dbarchdir = config.dbarchdir 
	reportdir = config.reportdir 
	tmpdir    = config.tmpdir
	bindir    = config.bindir
	confdir   = config.confdir

	for d in [basedir,dbdir,dbarchdir,tmpdir,bindir,confdir,reportdir]:
		if not os.path.isdir(d):
			cmd = 'mkdir -p %s' % d
			(stdout,stdin) = popen2.popen4(cmd)
			lns = stdout.readlines()
			if not os.path.isdir(d):
				msg = 'Unable to create directory: %s\n' % d
				sys.stderr.write(msg)
				for ln in lns:
					sys.stderr.write(ln)
				sys.exit(2)




def set_host_config():

	if options.verbose:
		sys.stderr.write('set_host_config()\n')

	# add config module to host objects so they can see the params there
	for h in config.host_list:
		h.config = config
		h.config.verbose = options.verbose
		h.ts_format = '%Y%m%d_%H.%M.%S'


def create_lockfile():
	if options.verbose:
		sys.stderr.write( 'create_lockfile()\n')

	lockfile = config.lockfile

	if os.path.exists(lockfile):

		if lockfile_is_stale(lockfile):
			msg = 'Removing stale lockfile: %s\n' % lockfile
			sys.stderr.write(msg)
			remove_lockfile()
		else:
			msg = 'Not running with valid lockfile in place: %s\n' % lockfile
			sys.stderr.write(msg)
			ts = time.strftime('%Y%m%d_%H.%M.%S')
			if options.verbose:
				sys.stderr.write( " - adding our pid to lockfile record\n")
				sys.stderr.write( " - open(%s,'a')\n" % lockfile)
			fh = open(lockfile,'a')
			fh.write('%s\n' % ts)
			sys.exit(0)

	lockdir = os.path.dirname(lockfile)
	if not os.path.exists(lockdir):
		# try to create directory
		if options.verbose:
			sys.stderr.write( " - os.system('mkdir -p %s')\n" % lockdir)
		os.system('mkdir -p %s' % lockdir)
		if not os.path.exists(lockdir):
			msg = "lockdir does not exist, can't be created: %s\n" % lockdir
			sys.stderr.write(msg)
			sys.exit(2)
	if options.verbose:
		sys.stderr.write(" - creating new pid/lockfile\n")
		sys.stderr.write(" - open('%s','w')\n" % lockfile)
	fh = open(lockfile,'w')
	pid = os.getpid()
	fh.write('%s\n' % pid)
	fh.close()


def remove_lockfile():
	if options.verbose:
		sys.stderr.write('remove_lockfile()\n')
		sys.stderr.write(' - os.remove(%s)\n' % config.lockfile)
	if config.lockfile:
		os.remove(config.lockfile)


def lockfile_is_stale(lockfile):
	'Returns false if there is a unix process with PID of lockfile, true otherwise'

	# works on solaris & linux... ymmv on other flavors

	fh = open(lockfile,'r')
	ln = fh.readline()

	cmd = 'ps -ef'
	pidre = re.compile(r'\s*(\d+)\s*')
	pspidre = re.compile(r'\s*\S+\s+(\d+)\s+')
	m = pidre.search(ln)
	if m:
		lockpid = m.group(1)
		(stdout,stdin) = popen2.popen4(cmd)
		output = stdout.readlines()
		for ln in output:
			m = pspidre.search(ln)
			if m:
				pspid = m.group(1)
				if pspid == lockpid:
					return False
	return True


def import_config(filename):
	'imports module located at <filename> and imports as global var "config"'

	if options.verbose:
		sys.stderr.write('import_config(%s)\n' % filename)
	global config

	if not os.path.exists(filename):
		msg = 'Regular file does not exist: %s\n' % filename
		sys.stderr.write(msg)
		sys.exit(1)

	dirname  = os.path.dirname(filename)
	basename = os.path.basename(filename)

	if basename[-3:] == '.py':
		modname = basename[0:-3]
	elif basename[-4:] == '.pyc':
		modname = basename[0:-4]
	else:
		modname = basename

	path = [dirname]

	(file,path,descr) = imp.find_module(modname,path)
	config = imp.load_module(modname,file,path,descr)


def	gen_host_reports():

	if options.verbose: 
		sys.stderr.write('gen_host_reports()\n')

	host_thread_list = []

	for h in config.host_list:

		host_thread = HostThread(h)
		host_thread.start()
		host_thread_list.append(host_thread)

		# sleep so we launch them at the predetermined rate
		time.sleep(config.popen_interval)

	for host_thread in host_thread_list:
		host_thread.join()


def gen_unified_report():

	ts = time.strftime('%Y%m%d_%H.%M.%S')

	import DarcReport, TextReport, HtmlReport, Sendmail

	textfile = '%s/darc_report_%s.txt'  % (config.reportdirbase,ts)
	htmlfile = '%s/darc_report_%s.html' % (config.reportdirbase,ts)

	textrep = TextReport.TextReport(config.host_list)
	textout = textrep.gen_report()

	fh = open(textfile,'w')
	for ln in textout:
		fh.write(ln)
	fh.close()

	htmlrep = HtmlReport.HtmlReport(config.host_list)
	htmlout = htmlrep.gen_report()

	fh = open(htmlfile,'w')
	for ln in htmlout:
		fh.write(ln)
	fh.close()

	if options.report_by_email:

		r = DarcReport.DarcReport(config.host_list)

		subject = '[darc report] %d failed, %d unchecked, %d passed' % (r.fail_c,r.prob_c,r.pass_c)

		from_addr = config.mail_from
		Sendmail.sendmail(config.rcpt_list,from_addr,subject,subject,[htmlfile])


def send_syslog_msg(msg):
	syslog.openlog('aide',0,config.syslog_facility)
	syslog.syslog(config.syslog_priority,msg)


def report_syslog():
	for h in config.host_list:
		if h.status   == 'FAIL':
			send_syslog_msg('host %s failed check' % h.hostname)
		elif h.status == 'PROB':
			send_syslog_msg('unable to check host %s' % h.hostname)


def init_host():
	found = False
	for x in config.host_list:
		if x.hostname == options.inithost:
			found = True
			x.main(init=True)
	if not found:
		msg = 'Unable to find host "%s" in config.host_list\n' % options.inithost
		sys.stderr.write(msg)
		sys.exit(1)


def test_host():
	found = False
	for x in config.host_list:
		if x.hostname == options.testhost:
			found = True
			x.test()
	if not found:
		msg = 'Unable to find host "%s" in config.host_list\n' % options.testhost
		sys.stderr.write(msg)
		sys.exit(1)
	

def find_last_archdb(h):
	'Returns path to most recent archived database file, based on directory/filename timestamps.'

	fnamere = re.compile(r'^(\S+)\.\d{8}_\d\d\.\d\d\.\d\d\.db')

	path = config.dbarchdir
	
	# add the yyyy-mm part
	lst = os.listdir(config.dbarchdir)
	lst.sort()
	path += '/' + str(lst[-1])

	# add the dd part
	lst = os.listdir(path)
	if len(lst) == 0:
		return False
	lst.sort()
	path += '/' + str(lst[-1])

	# add the filename part
	flist = []
	lst = os.listdir(path)
	for fname in lst:
		m = fnamere.search(fname)
		if m:
			if m.group(1) == h:
				flist.append(fname)
	
	if len(flist) == 0:
		return False
	else:
		flist.sort()
		path += '/%s' % flist[-1]
		return path



def update_host():
	h = options.update_host

	baseline = '%s/%s.db' % (config.dbdir,h)

	lastdb = find_last_archdb(h)

	if not lastdb:
		msg = '''Unable to locate latest archive file.  Verify expected directory structure and database file exists:  
%s/YYYY-MM/DD/hostname.YYYYMMDD_HH.mm.ss.db(.gz)
Alternatively, you can just copy whatever database you choose to 
%s/%s.db
''' % (config.dbdir,config.dbdir,h)
		sys.stderr.write(msg)
		sys.exit(2)

	if lastdb[-3:] == '.gz':
		cmd = 'gunzip -c '
	else:
		cmd = 'cat '
	
	cmd += ' %s > %s ' % (lastdb,baseline)

	sys.stderr.write('Updating %s\'s baseline from %s\n' % (h,lastdb))
	p = popen2.Popen4(cmd)
	code = p.wait()

	if code != 0:
		msg = 'Error!\n'
		sys.stderr.write(msg)
		errlns = p.fromchild.readlines()
		for ln in errlns:
			sys.stderr.write(ln)

def redefine_host_list():
	sh = None
	for h in config.host_list:
		if h.hostname == options.single_host:
			sh = h

	if sh:
		config.host_list = [sh]
	else:
		sys.stderr.write('Unable to find host %s in config.host_list\n')
		sys.exit(2)


def main():
	parse_opts()
	import_config(options.config)

	if options.update_host:
		update_host()
		sys.exit(0)

	add_dir_timestamps()
	mkdirs()
	set_host_config()

	# we run mkdirs() every time anyway in order to create the dated directories
	# but if the user specifies -k that's all we do.
	if options.mkdirs:
		sys.exit(0)

	if options.testhost:
		test_host()
	elif options.inithost:
		create_lockfile()
		init_host()
		remove_lockfile()
	else:
		create_lockfile()
		if options.single_host:
			redefine_host_list()
		gen_host_reports()
		gen_unified_report()
		if options.syslog:
			report_syslog()
		remove_lockfile()


if __name__ == '__main__':

	main()
